import bloqade.compiler.rewrite.common.canonicalize as canonicalize

from bloqade.ir import scalar
from bloqade.ir.control import waveform
from bloqade import var, cast
from decimal import Decimal


class TestUtils:
    def test_is_literal(self):
        assert canonicalize.is_literal(cast(1))
        assert not canonicalize.is_literal(var("a"))

    def test_is_negative(self):
        assert canonicalize.is_negative(-var("a"))
        assert not canonicalize.is_negative(var("a"))
        # canonicalize will put negative node into literals
        assert not canonicalize.is_negative(-cast(1))

    def test_is_zero(self):
        assert canonicalize.is_zero(cast(0))
        assert not canonicalize.is_zero(cast(1))
        assert not canonicalize.is_zero(var("a"))

    def test_is_one(self):
        assert canonicalize.is_one(cast(1))
        assert not canonicalize.is_one(cast(0))
        assert not canonicalize.is_one(var("a"))

    def test_is_scaled_waveform(self):
        assert canonicalize.is_scaled_waveform(cast("a") * waveform.Linear(0, 1, 1))

    def test_is_constant_waveform(self):
        assert canonicalize.is_constant_waveform(waveform.Constant(1, 1))


class TestScalar:
    def test_add(self):
        assert var("a") + var("b") == scalar.Add(var("a"), var("b"))
        assert var("a") + 1 == scalar.Add(var("a"), scalar.Literal(Decimal("1")))
        assert 0 + var("a") == var("a")
        assert var("a") + 0 == var("a")
        assert 1 + var("a") == scalar.Add(scalar.Literal(Decimal("1")), var("a"))
        assert cast(1) + cast(2) == scalar.Literal(Decimal("3"))
        assert var("a") - var("b") == scalar.Add(var("a"), scalar.Negative(var("b")))
        assert var("a") - 1 == scalar.Add(var("a"), cast(-1))
        assert cast(1) - cast(2) == scalar.Literal(Decimal("-1"))
        assert -cast(1) + cast(2) == scalar.Literal(Decimal("1"))
        assert 1 - var("a") == scalar.Add(
            scalar.Literal(Decimal("1")), scalar.Negative(var("a"))
        )
        assert -cast("a") + -cast("b") == scalar.Negative(
            expr=scalar.Add(var("a"), var("b"))
        )
        assert var("a") - var("a") == cast(0)
        assert -var("a") + var("a") == cast(0)

    def test_mul(self):
        assert var("a") * var("b") == scalar.Mul(var("a"), var("b"))
        assert 1 * var("a") == var("a")
        assert var("a") * 1 == var("a")
        assert 3 * var("a") == scalar.Mul(scalar.Literal(Decimal("3")), var("a"))
        assert cast(1) * cast(2) == scalar.Literal(Decimal("2"))
        assert -cast("a") * -cast("b") == cast("a") * cast("b")
        assert -cast("a") * cast("b") == -(cast("a") * cast("b"))
        assert cast("a") * -cast("b") == -(cast("a") * cast("b"))
        assert cast(0) * cast("a") == cast(0)
        assert cast(2) * cast(2) == cast(4)

    def test_div(self):
        assert var("a") / var("b") == scalar.Div(var("a"), var("b"))
        assert 1 / var("a") == scalar.Div(cast(1), var("a"))
        assert var("a") / 1 == var("a")
        assert 3 / var("a") == scalar.Div(cast(3), var("a"))
        assert cast(1) / cast(2) == cast(0.5)
        assert cast(0) / cast("a") == cast(0)
        assert -cast("a") / -cast("b") == cast("a") / cast("b")
        assert -cast("a") / cast("b") == -(cast("a") / cast("b"))
        assert cast("a") / -cast("b") == -(cast("a") / cast("b"))

    def test_negative_node(self):
        assert -cast(1.0) == cast(-1.0)
        assert -(-var("a")) == var("a")

    def test_min(self):
        a = var("a")
        b = var("b")
        c = var("c")
        d = cast(1)
        e = cast(2)
        f = cast(3)

        expr1 = a.min(b).min(c)
        expr2 = a.min(d).min(b).min(e).min(c).min(f)

        assert expr1 == scalar.Min([a, b, c])
        assert expr2 == scalar.Min([a, b, c, d])
        assert d.min(e) == d

    def test_max(self):
        a = var("a")
        b = var("b")
        c = var("c")
        d = cast(1)
        e = cast(2)
        f = cast(3)

        expr1 = a.max(b).max(c)
        expr2 = a.max(d).max(b).max(e).max(c).max(f)

        assert expr1 == scalar.Max([a, b, c])
        assert expr2 == scalar.Max([a, b, c, f])
        assert d.max(e) == e


class TestWaveform:
    def test_add(self):
        wf1 = waveform.Linear(0, "a", "T")
        wf2 = waveform.Constant(1, 1)
        wf3 = waveform.Constant(2, 1)
        wf4 = waveform.Poly(["a", "b", "c"], "t")

        assert wf1 + wf1 == 2 * wf1
        assert wf1 + waveform.Constant(1, 0) == wf1
        assert waveform.Constant(1, 0) + wf1 == wf1
        assert wf1 + waveform.Constant(0, 1) == wf1
        assert waveform.Constant(0, 1) + wf1 == wf1
        assert wf2 + wf3 == waveform.Constant(3, 1)
        assert 3.12 * wf1 + 3.12 * wf4 == 3.12 * (wf1 + wf4)
        assert 4.123 * wf1 + 3.13 * wf1 == 7.253 * wf1
        assert 4.123 * wf1 + wf1 == 5.123 * wf1
        assert wf1 + 4.123 * wf1 == 5.123 * wf1

    def test_neg(self):
        wf1 = waveform.Linear(0, "a", "T")
        wf2 = waveform.Constant(1, 1)

        assert -(-wf1) == wf1
        assert -wf2 == waveform.Constant(-1, 1)
        assert -(-var("a") * wf1) == var("a") * wf1

    def test_scale(self):
        wf1 = waveform.Linear(0, "a", "T")
        wf2 = waveform.Constant(1, 1)

        assert 1 * wf1 == wf1
        assert 0 * wf1 == waveform.Constant(0, "T")
        assert 2 * (3 * wf1) == 6 * wf1
        assert 2 * wf2 == waveform.Constant(2, 1)

    def test_append(self):
        # test nested appends

        wf1 = waveform.Append(
            [
                waveform.Constant(1, 1),
                waveform.Constant(2, 1),
                waveform.Constant(3, 1),
                waveform.Append([waveform.Constant(4, 1), waveform.Constant(5, 1)]),
            ]
        )
        canonicalize.Canonicalizer().visit(wf1) == waveform.Append(
            [
                waveform.Constant(1, 1),
                waveform.Constant(2, 1),
                waveform.Constant(3, 1),
                waveform.Constant(4, 1),
                waveform.Constant(5, 1),
            ]
        )

        wf2 = waveform.Append(
            [
                waveform.Constant(1, 1),
                waveform.Constant(2, 1),
                waveform.Constant(3, 1),
                waveform.Append(
                    [
                        waveform.Constant(4, 1),
                        waveform.Constant(5, 1),
                        waveform.Append(
                            [
                                waveform.Constant(6, 1),
                                waveform.Constant(7, 1),
                            ]
                        ),
                    ]
                ),
            ]
        )
        assert canonicalize.Canonicalizer().visit(wf2) == waveform.Append(
            [
                waveform.Constant(1, 1),
                waveform.Constant(2, 1),
                waveform.Constant(3, 1),
                waveform.Constant(4, 1),
                waveform.Constant(5, 1),
                waveform.Constant(6, 1),
                waveform.Constant(7, 1),
            ]
        )

        wf3 = waveform.Append(
            [
                waveform.Constant(1, 1),
                waveform.Constant(2, 1),
                waveform.Constant(3, 1),
                waveform.Append(
                    [
                        waveform.Constant(3, 1),
                        waveform.Constant(4, 0),
                    ]
                ),
            ]
        )
        # test that the constants are merged
        assert canonicalize.Canonicalizer().visit(wf3) == waveform.Append(
            [
                waveform.Constant(1, 1),
                waveform.Constant(2, 1),
                waveform.Constant(3, 2),
            ]
        )

    def test_slice(self):
        wf1 = waveform.Linear(0, "a", "T")

        T = var("T")

        sliced_wf1 = waveform.Slice(wf1, waveform.Interval(cast(0), T))

        assert canonicalize.Canonicalizer().visit(sliced_wf1) == wf1

        interval = waveform.Interval(var("a"), var("b"))

        assert canonicalize.Canonicalizer().visit(
            waveform.Slice(2.123 * wf1, interval)
        ) == 2.123 * waveform.Slice(wf1, interval)

        assert canonicalize.Canonicalizer().visit(
            waveform.Slice(-wf1, interval)
        ) == waveform.Negative(waveform.Slice(wf1, interval))

    def test_negative(self):
        wf1 = waveform.Linear(0, "a", "T")

        assert -(-wf1) == wf1
        assert -waveform.Constant(1, 1) == waveform.Constant(-1, 1)
        assert -waveform.Constant(-1, 1) == waveform.Constant(1, 1)


class TestHigherExpressions:
    def test_field(self):
        pass

    def test_pulse_leaf(self):
        pass

    def test_pulse_slice(self):
        pass

    def test_pulse_append(self):
        pass

    def test_sequence(self):
        pass

    def test_sequence_slice(self):
        pass

    def test_sequence_append(self):
        pass



import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
from bloqade.ir.visitor import BloqadeIRTransformer


class FlattenCircuit(BloqadeIRTransformer):
    # every visitor for sequence returns a Sequence

    def __init__(
        self,
        level_couplings=None,
        field_names=None,
        spatial_modulations=None,
    ):
        self.level_couplings = level_couplings
        self.field_names = field_names
        self.spatial_modulations = spatial_modulations
        self.duration = None

    #######################
    # Visitor definitions #
    #######################

    def visit_sequence_Append(self, node: sequence.Append) -> sequence.Sequence:
        seqs = list(map(self.visit, node.sequences))

        pulses = {}
        for lc in self.level_couplings:
            for s in seqs:
                p = s.pulses[lc]
                pulses[lc] = pulses[lc].append(p) if lc in pulses else p

        return self.visit(sequence.Sequence(pulses))

    def visit_sequence_Slice(self, node: sequence.Slice) -> sequence.Sequence:
        seq = self.visit(node.sequence)
        interval = node.interval

        pulses = {}

        for lc in self.level_couplings:
            p = seq.pulses[lc]
            pulses[lc] = p[interval.start : interval.stop]

        return self.visit(sequence.Sequence(pulses))

    def visit_sequence_NamedSequence(
        self, node: sequence.NamedSequence
    ) -> sequence.Sequence:
        return self.visit(node.sequence)

    def visit_pulse_Slice(self, node: pulse.Slice) -> pulse.Pulse:
        p = self.visit(node.pulse)
        interval = node.interval

        fields = {}

        for fn, f in p.fields.items():
            drives = {}
            for sm, wf in f.drives.items():
                drives[sm] = wf[interval.start : interval.stop]

            fields[fn] = field.Field(drives)

        return pulse.Pulse(fields)

    def visit_pulse_Append(self, node: pulse.Append) -> pulse.Pulse:
        pulses = list(map(self.visit, node.pulses))
        fields = dict(pulses[0].fields)

        for p in pulses[1:]:
            for fn, f in p.fields.items():
                for sm, wf in f.drives.items():
                    curr_wf = fields[fn].drives[sm]
                    fields[fn].drives[sm] = curr_wf.append(wf)

        return pulse.Pulse(fields)

    def visit_pulse_NamedPulse(self, node: pulse.NamedPulse) -> pulse.Pulse:
        return self.visit(node)



import bloqade.ir.control.waveform as _waveform
import bloqade.ir.control.field as _field
import bloqade.ir.control.pulse as _pulse
import bloqade.ir.control.sequence as _sequence
import bloqade.ir.location.location as _location
import bloqade.ir.analog_circuit as _analog_circuit
import bloqade.ir.scalar as _scalar
from beartype.typing import Any

from dataclasses import fields

BloqadeNodeTypes = (
    _scalar.Scalar,
    _scalar.Interval,
    _waveform.Waveform,
    _field.FieldExpr,
    _pulse.PulseExpr,
    _pulse.FieldName,
    _sequence.SequenceExpr,
    _sequence.LevelCoupling,
    _location.AtomArrangement,
    _location.ParallelRegister,
    _location.LocationInfo,
    _analog_circuit.AnalogCircuit,
)


# use dataclasses.fields to get the fields of a dataclass
def iter_fields(node: Any) -> Any:
    for field in fields(node):
        yield field, getattr(node, field.name)


class BloqadeIRVisitor:
    # Following the pattern from from python's node.NodeVisitor

    def visit(self, node: Any) -> Any:
        name = node.__class__.__name__
        module = node.__class__.__module__.split(".")[-1]
        visitor = getattr(self, f"visit_{module}_{name}", self.generic_visit)
        return visitor(node)

    def generic_visit(self, node: Any) -> Any:
        for field, value in iter_fields(node):
            if isinstance(value, BloqadeNodeTypes):
                self.visit(value)
            elif isinstance(value, (list, set, tuple, frozenset)):
                for item in value:
                    if isinstance(item, BloqadeNodeTypes):
                        self.visit(item)
            elif isinstance(value, dict):
                for key, value in value.items():
                    if isinstance(value, BloqadeNodeTypes):
                        self.visit(value)

                    if isinstance(key, BloqadeNodeTypes):
                        self.visit(key)


class BloqadeIRTransformer(BloqadeIRVisitor):
    # Following the pattern from from python's node.NodeTransformer

    def generic_visit(self, node: Any) -> Any:
        constructor_args = {}
        for field, value in iter_fields(node):
            if isinstance(value, BloqadeNodeTypes):
                new_value = self.visit(value)
            elif isinstance(value, (list, set, tuple, frozenset)):
                new_value = value.__class__(
                    (
                        self.visit(item) if isinstance(item, BloqadeNodeTypes) else item
                        for item in value
                    )
                )
            elif isinstance(value, dict):  # sometimes keys are also nodes
                new_value = {}
                for key, value in value.items():
                    if isinstance(value, BloqadeNodeTypes):
                        value = self.visit(value)
                    if isinstance(key, BloqadeNodeTypes):
                        key = self.visit(key)

                    new_value[key] = value
            else:
                new_value = value

            constructor_args[field.name] = new_value
        # IR nodes are immutable, so we have to use the constructor
        return node.__class__(**constructor_args)



from .add_padding import AddPadding
from .assign_to_literal import AssignToLiteral
from .assign_variables import AssignBloqadeIR
from .canonicalize import Canonicalizer
from .flatten import FlattenCircuit

__all__ = [
    "AddPadding",
    "AssignToLiteral",
    "AssignBloqadeIR",
    "Canonicalizer",
    "FlattenCircuit",
]



# from bloqade import start
from bloqade.ir import location, cast, Interpolation
from bloqade.ir import Constant, Linear, Poly, PythonFn, Sample
import pytest
import json
import numpy as np


def fc(x):
    return float(x)


def fvec(v):
    return list(map(fc, v))


def test_integration_jump_err():
    ## jump at the end of linear -- constant
    with pytest.raises(ValueError):
        (
            location.Square(6)
            .rydberg.detuning.uniform.apply(
                Constant("initial_detuning", "up_time")
                .append(Linear("initial_detuning", "final_detuning", "anneal_time"))
                .append(0.5 * Constant("final_detuning", "up_time"))
            )
            .assign(
                initial_detuning=-10,
                up_time=0.1,
                final_detuning=15,
                anneal_time=10,
            )
            .quera.mock()
            ._compile(10)
        )


def test_integration_scale():
    seq = Linear(start=0.0, stop=1.0, duration=0.5).append(
        2 * Constant(0.5, duration=0.5)
    )
    job = (
        location.Square(1)
        .rydberg.detuning.uniform.apply(seq)
        .quera.mock()
        ._compile(shots=10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 1, 1]) * 1e6)


def test_integration_neg():
    seq = Linear(start=0.0, stop=-0.5, duration=0.5).append(
        -Constant(0.5, duration=0.5)
    )
    job = (
        location.Square(1)
        .rydberg.detuning.uniform.apply(seq)
        .quera.mock()
        ._compile(shots=10)
    )

    panel = json.loads(job.json())
    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, -0.5, -0.5]) * 1e6)


def test_integration_poly_order_err():
    ## poly
    with pytest.raises(ValueError):
        seq = Poly(coeffs=[1, 2, 3], duration=0.5).append(-Constant(0.5, duration=0.5))
        (
            location.Square(1)
            .rydberg.detuning.uniform.apply(seq)
            .quera.mock()
            ._compile(shots=10)
        )


def test_integration_poly_const():
    ## constant
    seq = Poly(coeffs=[1], duration=0.5).append(Constant(1, duration=0.5))
    job = (
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(12)
    )

    panel = json.loads(job.json())
    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 12
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([1, 1, 1]) * 1e6)


def test_integration_poly_linear():
    ## linear
    seq = Poly(coeffs=[1, 2], duration=0.5).append(Constant(2, duration=0.5))
    job = (
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)
    )

    panel = json.loads(job.json())
    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([1, 2, 2]) * 1e6)


def test_integration_linear_sampl_const_err():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    wv = PythonFn.create(my_cos, duration=1.0)
    dt = cast(0.1)

    wf = Sample(wv, Interpolation.Constant, dt)
    ## phase can only have piecewise constant.
    with pytest.raises(ValueError):
        location.Square(1).rydberg.detuning.uniform.apply(wf).quera.mock()._compile(10)


def test_integration_slice_linear_const():
    seq = Linear(start=0.0, stop=1.0, duration=1.0)[0:0.5].append(
        Constant(0.5, duration=0.5)
    )
    job = (
        location.Square(1)
        .rydberg.detuning.uniform.apply(seq)
        .quera.mock()
        ._compile(shots=10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 0.5, 0.5]) * 1e6)


def test_integration_slice_linear_no_stop():
    seq = Linear(start=0.0, stop=1.0, duration=1.0)[0:]
    job = (
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 1.0]) * 1e6)


def test_integration_slice_linear_no_start():
    seq = Linear(start=0.0, stop=1.0, duration=1.0)[:1.0]
    job = (
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]
    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(detune_ir["global"]["values"]) == np.array([0, 1.0]) * 1e6)


def test_integration_slice_linear_error_neg_start():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=1.0)[-0.1:1.0]
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)


def test_integration_slice_linear_error_exceed_stop():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=1.0)[:4.0]
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)


def test_integration_slice_linear_error_revese():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=1.0)[0.5:0.0]
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)


def test_integration_slice_linear_error_same_vals_nofield():
    with pytest.raises(ValueError):
        seq = Linear(start=0.0, stop=1.0, duration=1.0)[0.0:0.0]
        location.Square(1).rydberg.detuning.uniform.apply(seq).quera.mock()._compile(10)


def test_integration_phase():
    batch = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.piecewise_constant(
            durations=[0.5, 0.5], values=[0, 1]
        )
        .record("a")
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(batch.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([0, 1.0, 1.0]))


def test_intergration_phase_nonconst_err():
    with pytest.raises(ValueError):
        (
            location.Square(1)
            .rydberg.rabi.phase.uniform.piecewise_linear(
                durations=[0.5, 0.5], values=[0, 1, 2]
            )
            .quera.mock()
            ._compile(10)
        )

    seq = Poly(coeffs=[1, 2], duration=0.5)
    with pytest.raises(ValueError):
        location.Square(1).rydberg.rabi.phase.uniform.apply(seq).quera.mock()._compile(
            10
        )


def test_integration_phase_linear():
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.linear(start=1, stop=1, duration=1)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_polyconst():
    seq = Poly(coeffs=[1], duration=0.5)
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_slice():
    ##[Further investigate!]
    seq = Poly(coeffs=[1], duration=1.0)[0:0.5]
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_scale():
    seq = 3.0 * Constant(value=1.0, duration=1.0)
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([3.0, 3.0]))


def test_integration_phase_neg():
    seq = -Constant(value=1.0, duration=1.0)
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 1.0]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([-1.0, -1.0]))


def test_integration_phase_slice_no_start():
    seq = Poly(coeffs=[1], duration=1.0)[:0.5]
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_slice_no_stop():
    ##[Further investigate!]
    seq = Poly(coeffs=[1], duration=0.5)[0:]
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_slice_same_start_stop():
    ##[Further investigate!]
    seq = Poly(coeffs=[1], duration=0.5)[0:0.5]
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.apply(seq)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([1.0, 1.0]))


def test_integration_phase_slice_error_neg_start():
    with pytest.raises(ValueError):
        seq = Poly(coeffs=[1], duration=1.0)[-0.1:0.5]
        location.Square(1).rydberg.rabi.phase.uniform.apply(seq).quera.mock()._compile(
            10
        )


def test_integration_phase_slice_error_exceed_stop():
    with pytest.raises(ValueError):
        seq = Poly(coeffs=[1], duration=1.0)[0:2.0]
        location.Square(1).rydberg.rabi.phase.uniform.apply(seq).quera.mock()._compile(
            10
        )


def test_integration_phase_slice_error_reverse():
    with pytest.raises(ValueError):
        seq = Poly(coeffs=[1], duration=1.0)[2.0:0.0]
        location.Square(1).rydberg.rabi.phase.uniform.apply(seq).quera.mock()._compile(
            10
        )


def test_integration_phase_sampl_linear_err():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    wv = PythonFn.create(my_cos, duration=1.0)
    dt = cast(0.1)

    wf = Sample(wv, Interpolation.Linear, dt)
    ## phase can only have piecewise constant.
    with pytest.raises(ValueError):
        location.Square(1).rydberg.rabi.phase.uniform.apply(wf).quera.mock()._compile(
            10
        )


def test_integration_batchassign_assign():
    ## jump at the end of linear -- constant
    with pytest.raises(ValueError):
        (
            location.Square(6)
            .rydberg.detuning.uniform.apply(Constant("initial_detuning", "time"))
            .batch_assign(initial_detuning=[1, 2, 3, 4], time=[2, 4, 5])
            .quera.mock()
            ._compile(10)
        )

    ## jump at the end of linear -- constant

    batch = (
        location.Square(6)
        .rydberg.detuning.uniform.apply(Constant("initial_detuning", "time"))
        .batch_assign(initial_detuning=[1, 2, 3, 4], time=[2, 4, 5, 0.6])
        .quera.mock()
        ._compile(10)
    )

    assert len(batch.tasks) == 4


def test_integration_record():
    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.piecewise_constant(
            durations=[0.5, 0.5], values=[0, 1]
        )
        .record("a")
        .piecewise_constant(durations=[0.3], values=["a"])
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5, 1.3]) * 1e-6)
    assert all(fvec(phase_ir["global"]["values"]) == np.array([0, 1.0, 1.0]))


pytest.mark.skip(reason="fixing now")


def test_integration_fn_phase():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    job = (
        location.Square(1)
        .rydberg.rabi.phase.uniform.fn(my_cos, duration=1.0)
        .sample(dt=0.5)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    phase_ir = ir["effective_hamiltonian"]["rydberg"]["rabi_frequency_phase"]
    assert all(fvec(phase_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(
        fvec(phase_ir["global"]["values"]) == np.array([1.0, 0.8775825, 0.8775825])
    )


pytest.mark.skip(reason="fixing now")


def test_integration_fn_detune():
    def my_cos(time):
        return np.cos(time)

    assert my_cos(1) == np.cos(1)

    job = (
        location.Square(1)
        .rydberg.detuning.uniform.fn(my_cos, duration=1.0)
        .sample(dt=0.5)
        .quera.mock()
        ._compile(10)
    )

    panel = json.loads(job.json())

    print(panel)

    task_data = panel["bloqade.task.batch.RemoteBatch"]["tasks"][0][1]

    ir = task_data["bloqade.task.quera.QuEraTask"]["task_ir"]

    assert ir["nshots"] == 10
    assert fvec(ir["lattice"]["sites"][0]) == [0.0, 0.0]
    assert ir["lattice"]["filling"] == [1]
    assert ir["lattice"]["filling"] == [1]

    detune_ir = ir["effective_hamiltonian"]["rydberg"]["detuning"]
    assert all(fvec(detune_ir["global"]["times"]) == np.array([0, 0.5, 1.0]) * 1e-6)
    assert all(
        fvec(detune_ir["global"]["values"]) == np.array([1000000, 877582.6, 540302.4])
    )



import pytest
from bloqade import start
import numpy as np
from unittest.mock import patch
from bloqade.submission.ir.task_results import QuEraTaskStatusCode, QuEraTaskResults
from bloqade.submission.ir.task_specification import QuEraTaskSpecification
from bloqade.task.base import Geometry
from bloqade.task.quera import QuEraTask
from bloqade.task.braket import BraketTask
from bloqade.submission.base import ValidationError
from bloqade import dumps, loads

# import numpy as np


def mock_results_json(L):
    pre_sequence = np.array([0] + [1] * (L - 1))
    post_sequence = pre_sequence * np.random.choice(2, size=L)

    pre_sequence = pre_sequence.tolist()
    post_sequence = post_sequence.tolist()
    return {
        "task_status": "Completed",
        "shot_outputs": [
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
            {
                "shot_status": "Completed",
                "pre_sequence": pre_sequence,
                "post_sequence": post_sequence,
            },
        ],
    }


def mock_task_ir(L):
    return {
        "nshots": 10,
        "lattice": {"sites": L * [(0, 0)], "filling": L * [1]},
        "effective_hamiltonian": {
            "rydberg": {
                "rabi_frequency_amplitude": {
                    "global": {
                        "times": [0, 1e-6, 2e-6, 3e-6, 4e-6],
                        "values": [0, 15e6, 15e6, 0],
                    },
                },
                "rabi_frequency_phase": {
                    "global": {"times": [0, 4e-6], "values": [0, 0]},
                },
                "detuning": {
                    "global": {
                        "times": [0, 1e-6, 2e-6, 3e-6, 4e-6],
                        "values": [0, 15e6, 15e6, 0],
                    },
                },
            }
        },
    }


def mock_task(L):
    return QuEraTaskSpecification(**mock_task_ir(L))


def mock_results(L):
    return QuEraTaskResults(**mock_results_json(L))


def test_base_classes():
    from bloqade.task.base import Task, LocalTask, RemoteTask

    task = Task()

    with pytest.raises(NotImplementedError):
        task.geometry

    task = LocalTask()

    with pytest.raises(NotImplementedError):
        task.result()

    with pytest.raises(NotImplementedError):
        task.run()

    task = RemoteTask()

    with pytest.raises(NotImplementedError):
        task.validate()

    with pytest.raises(NotImplementedError):
        task.result()

    with pytest.raises(NotImplementedError):
        task.fetch()

    with pytest.raises(NotImplementedError):
        task.status()

    with pytest.raises(NotImplementedError):
        task.pull()

    with pytest.raises(NotImplementedError):
        task.cancel()

    with pytest.raises(NotImplementedError):
        task.submit(True)

    with pytest.raises(NotImplementedError):
        task._result_exists()


def test_metadata_filter_scalar():
    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10)
    )

    filtered_batch = batch.filter_metadata(d=[1, 2, 16])

    assert filtered_batch.tasks.keys() == {0, 1, 4}

    with pytest.raises(ValueError):
        filtered_batch = batch.filter_metadata(d=[1, 2, 16, 1j])


def test_metadata_filter_vector():
    batch = (
        start.add_position([(0, i * 6.1) for i in range(2)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.scale("m")
        .piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8], m=[[0, 1], [1, 0], [1, 1], [0, 0]])
        .bloqade.python()
        ._compile(10)
    )
    filters = dict(d=[1, 8], m=[[0, 1], [1, 0]])
    filtered_batch_all = batch.filter_metadata(False, **filters)
    filtered_batch_any = batch.filter_metadata(True, **filters)

    assert filtered_batch_all.tasks.keys() == {0}
    assert filtered_batch_any.tasks.keys() == {0, 1, 3}

    filters = dict(d=[1, 8], m=[[0, 1], [1, 0], (0, 0)])

    with pytest.raises(ValueError):
        filtered_batch_all = batch.filter_metadata(**filters)


@patch("bloqade.ir.routine.braket.BraketBackend")
@patch("bloqade.task.batch.np.random.permutation")
def test_remote_batch_task_metric(permutation, BraketBackend):
    backend = BraketBackend(
        device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    )
    permutation.return_value = [0, 1, 2, 3, 4, 5]

    backend.submit_task.side_effect = list(map(str, range(6)))

    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10)
    )

    for k, task in batch.tasks.items():
        task.backend = backend

    batch._submit(ignore_submission_error=True, shuffle_submit_order=True)

    assert str(batch) == (
        "  task ID    status  shots\n"
        "0       0  Enqueued     10\n"
        "1       1  Enqueued     10\n"
        "2       2  Enqueued     10\n"
        "3       3  Enqueued     10\n"
        "4       4  Enqueued     10\n"
        "5       5  Enqueued     10"
    )

    assert batch.total_nshots == 60


@patch("bloqade.ir.routine.braket.BraketBackend")
def test_pull(BraketBackend):
    backend = BraketBackend(
        device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    )

    backend.submit_task.side_effect = list(map(str, range(6)))
    backend.task_status.return_value = QuEraTaskStatusCode.Completed
    backend.task_results.return_value = mock_results(14)

    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10)
    )

    for k, task in batch.tasks.items():
        task.backend = backend

    batch._submit(ignore_submission_error=True, shuffle_submit_order=False)
    batch.pull()
    new_batch = batch.remove_invalid_tasks()
    assert len(new_batch.tasks) == 6


@patch("bloqade.ir.routine.braket.BraketBackend")
def test_retrieve(BraketBackend):
    backend = BraketBackend(
        device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    )

    backend.submit_task.side_effect = list(map(str, range(6)))
    backend.task_results.return_value = mock_results(14)
    backend.task_status.side_effect = [
        # calls for fetch
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Running,
        QuEraTaskStatusCode.Enqueued,
        QuEraTaskStatusCode.Enqueued,
        # calls for retrieve
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
    ]

    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10)
    )

    for k, task in batch.tasks.items():
        task.backend = backend

    batch._submit(ignore_submission_error=True, shuffle_submit_order=False)
    batch.fetch()
    batch.retrieve()
    new_batch = batch.remove_invalid_tasks()
    assert len(new_batch.tasks) == 6


@patch("bloqade.ir.routine.braket.BraketBackend")
def test_remove_invalid_tasks(BraketBackend):
    backend = BraketBackend(
        device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    )

    backend.submit_task.side_effect = list(map(str, range(6)))
    backend.task_status.return_value = QuEraTaskStatusCode.Unaccepted

    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10)
    )

    for k, task in batch.tasks.items():
        task.backend = backend

    batch._submit(ignore_submission_error=True, shuffle_submit_order=False)
    batch.fetch()
    new_batch = batch.remove_invalid_tasks()
    assert len(new_batch.tasks) == 0


@patch("bloqade.ir.routine.braket.BraketBackend")
def test_test_filters(BraketBackend):
    backend = BraketBackend(
        device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    )

    backend.submit_task.side_effect = list(map(str, range(6)))
    backend.task_status.side_effect = [
        QuEraTaskStatusCode.Failed,
        QuEraTaskStatusCode.Cancelled,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
    ]

    backend.task_results.return_value = mock_results(14)

    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10)
    )

    for k, task in batch.tasks.items():
        task.backend = backend

    batch._submit(ignore_submission_error=True, shuffle_submit_order=False)
    batch.fetch()

    new_batch = batch.get_completed_tasks()
    assert len(new_batch.tasks) == 4

    new_batch = batch.get_failed_tasks()
    assert len(new_batch.tasks) == 1

    new_batch = batch.remove_failed_tasks()
    assert len(new_batch.tasks) == 5

    new_batch = batch.get_finished_tasks()
    assert len(new_batch.tasks) == 6


@patch("bloqade.ir.routine.braket.BraketBackend")
def test_cancel_tasks(BraketBackend):
    backend = BraketBackend(
        device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    )

    backend.submit_task.side_effect = list(map(str, range(6)))
    backend.cancel_task.return_value = None

    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10)
    )

    for k, task in batch.tasks.items():
        task.backend = backend

    batch._submit(ignore_submission_error=True, shuffle_submit_order=False)
    batch.cancel()

    assert backend.cancel_task.call_count == 6


@patch("bloqade.ir.routine.braket.BraketBackend")
def test_resubmit_tasks(BraketBackend):
    backend = BraketBackend(
        device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    )

    backend.submit_task.side_effect = list(map(str, range(12)))

    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10, name="test")
    )

    for k, task in batch.tasks.items():
        task.backend = backend

    batch._submit(ignore_submission_error=True, shuffle_submit_order=False)
    batch.resubmit(shuffle_submit_order=False)
    print(batch)
    assert str(batch) == (
        "  task ID    status  shots\n"
        "0       6  Enqueued     10\n"
        "1       7  Enqueued     10\n"
        "2       8  Enqueued     10\n"
        "3       9  Enqueued     10\n"
        "4      10  Enqueued     10\n"
        "5      11  Enqueued     10"
    )


@patch("bloqade.ir.routine.braket.BraketBackend")
def test_report(BraketBackend):
    backend = BraketBackend(
        device_arn="arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
    )

    backend.submit_task.side_effect = list(map(str, range(6)))
    backend.task_status.side_effect = [
        QuEraTaskStatusCode.Failed,
        QuEraTaskStatusCode.Partial,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Completed,
        QuEraTaskStatusCode.Cancelled,
    ]

    backend.task_results.return_value = mock_results(14)

    batch = (
        start.add_position([(0, i * 6.1) for i in range(14)])
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 3.8, 0.1], [-20, -20, "d", "d"]
        )
        .amplitude.uniform.piecewise_linear([0.1, 3.8, 0.1], [0, 15, 15, 0])
        .phase.uniform.constant(np.pi / 2, 4)
        .batch_assign(d=[1, 2, 4, 8, 16, 32])
        .braket.aquila()
        ._compile(10)
    )

    for k, task in batch.tasks.items():
        task.backend = backend

    batch._submit(ignore_submission_error=True, shuffle_submit_order=False)
    batch.fetch()
    report = batch.report()

    assert report.list_param("d") == [2.0, 4.0, 8.0, 16.0]
    assert report.markdown == (report.dataframe.to_markdown())

    assert all([len(ele) == 0 for ele in report.bitstrings(True)])
    assert all([len(ele) == 10 for ele in report.bitstrings(False)])
    assert all([len(ele) == 0 for ele in report.counts()])


@patch("bloqade.task.quera.QuEraBackend")
def test_braket_task(QuEraBackend):
    backend = QuEraBackend()

    task = QuEraTask(
        task_id=None,
        backend=backend,
        task_ir=mock_task(14),
        metadata={},
    )

    assert task.nshots == 10
    assert task.geometry == Geometry([(0, 0)] * 14, [1] * 14)
    assert task.status() == QuEraTaskStatusCode.Unsubmitted
    assert not task._result_exists()

    with pytest.warns(Warning):
        task.cancel()

    with pytest.raises(ValueError):
        task.fetch()

    with pytest.raises(ValueError):
        task.pull()

    backend.submit_task.side_effect = ["1", "2"]

    task.submit()
    assert task.task_id == "1"

    with pytest.raises(ValueError):
        task.submit()

    task.submit(True)
    assert task.task_id == "2"

    backend.validate_task.return_value = None

    assert task.validate() == ""

    backend.validate_task.side_effect = ValidationError("test")
    assert task.validate() == "test"

    backend.task_status.return_value = QuEraTaskStatusCode.Completed
    mock_result = mock_results(14)
    backend.task_results.return_value = mock_result
    task.fetch()

    assert task.status() == QuEraTaskStatusCode.Completed
    assert task.result() == mock_result
    assert task._result_exists()
    assert task == task.fetch()

    task.cancel()
    assert backend.cancel_task.call_count == 1


@patch("bloqade.task.braket.BraketBackend")
def test_quera_task(BraketBackend):
    backend = BraketBackend()

    task = BraketTask(
        task_id=None,
        backend=backend,
        task_ir=mock_task(14),
        metadata={},
    )

    assert task.nshots == 10
    assert task.geometry == Geometry([(0, 0)] * 14, [1] * 14)
    assert task.status() == QuEraTaskStatusCode.Unsubmitted
    assert not task._result_exists()

    with pytest.warns(Warning):
        task.cancel()

    with pytest.raises(ValueError):
        task.fetch()

    with pytest.raises(ValueError):
        task.pull()

    backend.submit_task.side_effect = ["1", "2"]

    task.submit()
    assert task.task_id == "1"

    with pytest.raises(ValueError):
        task.submit()

    task.submit(True)
    assert task.task_id == "2"

    backend.validate_task.return_value = None

    assert task.validate() == ""

    backend.validate_task.side_effect = ValidationError("test")
    assert task.validate() == "test"

    backend.task_status.return_value = QuEraTaskStatusCode.Completed
    mock_result = mock_results(14)
    backend.task_results.return_value = mock_result
    task.fetch()

    assert task.status() == QuEraTaskStatusCode.Completed
    assert task.result() == mock_result
    assert task._result_exists()
    assert task == task.fetch()

    task.cancel()
    assert backend.cancel_task.call_count == 1


def test_serializer():
    prog = (
        start.add_position((0, 0))
        .rydberg.detuning.uniform.piecewise_linear(
            [0.1, 0.5, 0.1], [1.0, 2.0, 3.0, 4.0]
        )
        .constant(4.0, 1)
    )

    for backend in [
        "braket.aquila",
        "braket.local_emulator",
        "bloqade.python",
        "quera.aquila",
    ]:
        print(backend)
        batch = prog.device(backend)._compile(10)

        json_str = dumps(batch, use_decimal=True)
        new_batch = loads(json_str, use_decimal=True)
        for task_num, task in batch.tasks.items():
            if hasattr(task, "emulator_ir"):
                assert task.emulator_ir == new_batch.tasks[task_num].emulator_ir
            else:
                assert task.task_ir == new_batch.tasks[task_num].task_ir
            assert task.metadata == new_batch.tasks[task_num].metadata
            assert task.task_result_ir == new_batch.tasks[task_num].task_result_ir
            assert task.nshots == new_batch.tasks[task_num].nshots



from collections import OrderedDict

from beartype.typing import List, Sequence, Union, Dict, Optional, Tuple
from beartype import beartype
from numbers import Number

from bloqade.submission.ir.task_results import (
    QuEraTaskResults,
    QuEraTaskStatusCode,
)
from numpy.typing import NDArray
import pandas as pd
import numpy as np
from pydantic.dataclasses import dataclass
from bloqade.submission.ir.parallel import ParallelDecoder
import datetime
from bloqade.visualization import display_report


@dataclass(frozen=True)
class Geometry:
    sites: List[Tuple[float, float]]
    filling: List[int]
    parallel_decoder: Optional[ParallelDecoder] = None


class Task:
    def _geometry(self) -> Geometry:
        raise NotImplementedError(
            f"{self.__class__.__name__}._geometry() not implemented"
        )

    @property
    def geometry(self) -> Geometry:
        return self._geometry()


class RemoteTask(Task):
    def validate(self) -> None:
        raise NotImplementedError

    def result(self) -> QuEraTaskResults:
        # online, Blocking
        # waiting for remote results to finish
        # return results
        raise NotImplementedError

    def fetch(self) -> None:
        # online, non-blocking
        # pull the result if they are complete
        raise NotImplementedError

    def status(self) -> QuEraTaskStatusCode:
        # online, non-blocking
        # probe current task status
        raise NotImplementedError

    def pull(self) -> None:
        # online, blocking
        # pull the current results
        raise NotImplementedError

    def cancel(self) -> None:
        # this method cancels the task
        raise NotImplementedError

    def submit(self, force: bool):
        # online, non-blocking
        # this method submit the task
        raise NotImplementedError

    def _result_exists(self) -> bool:
        raise NotImplementedError


class LocalTask(Task):
    def result(self):
        # need a new results type
        # for emulator jobs
        raise NotImplementedError

    def run(self, **kwargs):
        raise NotImplementedError


# Report is now just a helper class for
# organize and analysis data:
class Report:
    dataframe: pd.DataFrame
    metas: List[Dict]
    geos: List[Geometry]
    name: str = ""

    def __init__(self, data, metas, geos, name="") -> None:
        self.dataframe = data  # df
        self._bitstrings = None  # bitstring cache
        self._counts = None  # counts cache
        self.metas = metas
        self.geos = geos
        self.name = name + " " + str(datetime.datetime.now())

    def list_param(self, field_name: str) -> List[Union[Number, None]]:
        """
        List the parameters associate with the given variable field_name
        for each tasks.

        Args:
            field_name (str): variable name

        """

        def cast(x):
            if x is None:
                return None
            elif isinstance(x, (list, tuple, np.ndarray)):
                return list(map(cast, x))
            else:
                return float(x)

        return list(map(cast, (meta.get(field_name) for meta in self.metas)))

    @property
    def markdown(self) -> str:
        return self.dataframe.to_markdown()

    def _filter(
        self,
        *,
        task_number: Optional[int] = None,
        filter_perfect_filling: bool = True,
        clusters: Union[tuple[int, int], Sequence[Tuple[int, int]]] = [],
    ):
        mask = np.ones(len(self.dataframe), dtype=bool)

        if task_number is not None:
            task_numbers = self.dataframe.index.get_level_values("task_number")
            np.logical_and(task_numbers == task_number, mask, out=mask)

        if filter_perfect_filling:
            perfect_sorting = self.dataframe.index.get_level_values("perfect_sorting")
            pre_sequence = self.dataframe.index.get_level_values("pre_sequence")

            np.logical_and(perfect_sorting == pre_sequence, mask, out=mask)

        clusters = [clusters] if isinstance(clusters, tuple) else clusters

        for cluster in clusters:
            given_clusters = self.dataframe.index.get_level_values("cluster")
            np.logical_and(given_clusters == cluster, mask, out=mask)

        return mask

    @beartype
    def bitstrings(
        self,
        filter_perfect_filling: bool = True,
        clusters: Union[tuple[int, int], List[tuple[int, int]]] = [],
    ) -> List[NDArray]:
        """Get the bitstrings from the data.

        Args:
            filter_perfect_filling (bool): whether return will
            only contain perfect filling shots. Defaults to True.
            clusters: (tuple[int, int], Sequence[Tuple[int, int]]):
            cluster index to filter shots from. If none are provided
            all clusters are used, defaults to [].

        Returns:
            bitstrings (list of ndarray): list corresponding to each
            task in the report. Each element is an ndarray of shape
            (nshots, nsites) where nshots is the number of shots for
            the task and nsites is the number of sites in the task.

        Note:
            Note that nshots may vary between tasks if filter_perfect_filling
            is set to True.

        """

        task_numbers = self.dataframe.index.get_level_values("task_number").unique()

        bitstrings = []
        for task_number in task_numbers:
            mask = self._filter(
                task_number=task_number,
                filter_perfect_filling=filter_perfect_filling,
                clusters=clusters,
            )
            if np.any(mask):
                bitstrings.append(self.dataframe.loc[mask].to_numpy())
            else:
                bitstrings.append(
                    np.zeros((0, self.dataframe.shape[1]), dtype=np.uint8)
                )

        return bitstrings

    @beartype
    def counts(
        self,
        filter_perfect_filling: bool = True,
        clusters: Union[tuple[int, int], List[tuple[int, int]]] = [],
    ) -> List[OrderedDict[str, int]]:
        """Get the counts of unique bit strings.

        Args:
            filter_perfect_filling (bool): whether return will
            only contain perfect filling shots. Defaults to True.
            clusters: (tuple[int, int], Sequence[Tuple[int, int]]):
            cluster index to filter shots from. If none are provided
            all clusters are used, defaults to [].

        Returns:
            bitstrings (list of ndarray): list corresponding to each
            task in the report. Each element is an ndarray of shape
            (nshots, nsites) where nshots is the number of shots for
            the task and nsites is the number of sites in the task.

        Note:
            Note that nshots may vary between tasks if filter_perfect_filling
            is set to True.

        """

        def generate_counts(bitstring):
            output = np.unique(bitstring, axis=0, return_counts=True)

            count_list = [
                ("".join(map(str, bitstring)), count)
                for bitstring, count in zip(*output)
            ]
            count_list.sort(key=lambda x: x[1], reverse=True)
            count = OrderedDict(count_list)

            return count

        return list(
            map(generate_counts, self.bitstrings(filter_perfect_filling, clusters))
        )

    @beartype
    def rydberg_densities(
        self,
        filter_perfect_filling: bool = True,
        clusters: Union[tuple[int, int], List[tuple[int, int]]] = [],
    ) -> Union[pd.Series, pd.DataFrame]:
        """Get rydberg density for each task.

        Args:
            filter_perfect_filling (bool, optional): whether return will
            only contain perfect filling shots. Defaults to True.

        Return:
            per-site rydberg density for each task as a pandas DataFrame or Series.

        """
        mask = self._filter(
            filter_perfect_filling=filter_perfect_filling, clusters=clusters
        )
        df = self.dataframe[mask]
        return 1 - (df.groupby("task_number").mean())

    def show(self):
        """
        Interactive Visualization of the Report

        """
        display_report(self)



from bloqade.builder.typing import ParamType

from bloqade.compiler.passes.hardware.components import AHSComponents
from bloqade.ir import analog_circuit
from bloqade.ir.control import pulse, sequence, field

from bloqade.submission.ir.braket import BraketTaskSpecification

from bloqade.submission.ir.capabilities import QuEraCapabilities
from bloqade.submission.ir.task_specification import QuEraTaskSpecification

from beartype.typing import Dict, Optional, Tuple


def analyze_channels(circuit: analog_circuit.AnalogCircuit) -> Dict:
    """1. Scan channels

    This pass checks to make sure that:
    * There is no hyperfine coupling in the sequence
    * There are no non-uniform spatial modulation for rabi phase and amplitude
    * there is no more than one non-uniform spatial modulation for detuning

    Args:
        circuit: AnalogCircuit to analyze

    Returns:
        level_couplings: Dictionary containing the required channels for the
            sequence. Note that this will insert a uniform field for any missing
            channels.

    Raises:
        ValueError: If there is hyperfine coupling in the sequence.
        ValueError: If there is more than one non-uniform spatial modulation for
            detuning.
        ValueError: If there are non-uniform spatial modulations for rabi phase
            and amplitude.

    """
    from bloqade.compiler.analysis.hardware import ValidateChannels
    from bloqade.compiler.analysis.common import ScanChannels

    ValidateChannels().scan(circuit)
    level_couplings = ScanChannels().scan(circuit)

    # add missing channels
    fields = level_couplings[sequence.rydberg]
    # detuning, phase and amplitude are required
    # to have at least a uniform field
    updated_fields = {
        field_name: fields.get(field_name, {field.Uniform}).union({field.Uniform})
        for field_name in [pulse.detuning, pulse.rabi.amplitude, pulse.rabi.phase]
    }

    return {sequence.rydberg: updated_fields}


def canonicalize_circuit(
    circuit: analog_circuit.AnalogCircuit, level_couplings: Dict
) -> analog_circuit.AnalogCircuit:
    """2. Insert zero waveform in the explicit time intervals missing a waveform

    This pass inserts a zero waveform in the explicit time intervals missing a
    waveform. This is required for later analysis passes to check that the
    waveforms are compatible with the hardware.

    Args:
        circuit: AnalogCircuit to add padding to
        level_couplings: Dictionary containing the given channels for the
            sequence.

    Return
        circuit: AnalogCircuit with zero waveforms inserted in the explicit time
            intervals missing a waveform.

    """
    from bloqade.compiler.rewrite.common import (
        AddPadding,
        AssignToLiteral,
        Canonicalizer,
    )

    circuit = AddPadding(level_couplings).visit(circuit)
    # these two passes are equivalent to a constant propagation pass
    circuit = AssignToLiteral().visit(circuit)
    circuit = Canonicalizer().visit(circuit)

    return circuit


def assign_circuit(
    circuit: analog_circuit.AnalogCircuit, assignments: Dict[str, ParamType]
) -> Tuple[analog_circuit.AnalogCircuit, Dict]:
    """3. Assign variables and validate assignment

    This pass assigns variables to the circuit and validates that all variables
    have been assigned.

    Args:
        circuit: AnalogCircuit to assign variables to
        assignments: Dictionary containing the assignments for the variables in
            the circuit.

    Returns:
        assigned_circuit: AnalogCircuit with variables assigned.

    Raises:
        ValueError: If there are any variables that have not been assigned.

    """
    from bloqade.compiler.analysis.common import AssignmentScan, ScanVariables
    from bloqade.compiler.rewrite.common import AssignBloqadeIR

    final_assignments = AssignmentScan(assignments).scan(circuit)

    assigned_circuit = AssignBloqadeIR(final_assignments).visit(circuit)

    assignment_analysis = ScanVariables().scan(assigned_circuit)

    if not assignment_analysis.is_assigned:
        missing_vars = assignment_analysis.scalar_vars.union(
            assignment_analysis.vector_vars
        )
        raise ValueError(
            "Missing assignments for variables:\n"
            + ("\n".join(f"{var}" for var in missing_vars))
            + "\n"
        )

    return assigned_circuit, final_assignments


def validate_waveforms(
    level_couplings: Dict, circuit: analog_circuit.AnalogCircuit
) -> None:
    """4. validate piecewise linear and piecewise constant pieces of pulses

    This pass check to make sure that the waveforms are compatible with the
    hardware. This includes checking that the waveforms are piecewise linear or
    piecewise constant. It also checks that the waveforms are compatible with
    the given channels.

    Args:
        circuit: AnalogCircuit to validate waveforms for
        level_couplings: Dictionary containing the given channels for the
            sequence.

    Raises:
        ValueError: If the waveforms are not piecewise linear or piecewise
            constant, e.g. the waveform is not continuous.
        ValueError: If a waveform segment is not compatible with the given
            channels.

    """
    from bloqade.compiler.analysis.hardware import (
        ValidatePiecewiseConstantChannel,
        ValidatePiecewiseLinearChannel,
    )
    from bloqade.compiler.analysis.common import CheckSlices

    channel_iter = (
        (level_coupling, field_name, sm)
        for level_coupling, fields in level_couplings.items()
        for field_name, spatial_modulations in fields.items()
        for sm in spatial_modulations
    )
    for channel in channel_iter:
        if channel[1] in [pulse.detuning, pulse.rabi.amplitude]:
            ValidatePiecewiseLinearChannel(*channel).visit(circuit)
        else:
            ValidatePiecewiseConstantChannel(*channel).visit(circuit)

    CheckSlices().visit(circuit)

    if circuit.sequence.duration() == 0:
        raise ValueError("Circuit Duration must be be non-zero")


def generate_ahs_code(
    capabilities: Optional[QuEraCapabilities],
    level_couplings: Dict,
    circuit: analog_circuit.AnalogCircuit,
) -> AHSComponents:
    """5. generate ahs code

    Generates the AHS code for the given circuit. This includes generating the
    lattice data, global detuning, global amplitude, global phase, local
    detuning and lattice site coefficients (if applicable).

    Args:
        capabilities (QuEraCapabilities | None): Capabilities of the hardware.
        level_couplings (Dict): Dictionary containing the given channels for the
            sequence.
        circuit (AnalogCircuit): AnalogCircuit to generate AHS code for.

    Returns:
        ahs_components (AHSComponents): A collection of the AHS components
            generated for the given circuit. Can be used to generate the QuEra
            and Braket IR.

    Raises:
        ValueError: If the capabilities are not provided but the circuit has
            a ParallelRegister. This is because the ParallelRegister requires
            the capabilities to generate the lattice data.

    """
    from bloqade.compiler.codegen.hardware import (
        GenerateLattice,
        GenerateLatticeSiteCoefficients,
        GeneratePiecewiseLinearChannel,
        GeneratePiecewiseConstantChannel,
    )

    ahs_lattice_data = GenerateLattice(capabilities).emit(circuit)

    global_detuning = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.detuning, field.Uniform
    ).visit(circuit)

    global_amplitude = GeneratePiecewiseLinearChannel(
        sequence.rydberg, pulse.rabi.amplitude, field.Uniform
    ).visit(circuit)

    global_phase = GeneratePiecewiseConstantChannel(
        sequence.rydberg, pulse.rabi.phase, field.Uniform
    ).visit(circuit)

    local_detuning = None
    lattice_site_coefficients = None

    extra_sm = set(level_couplings[sequence.rydberg][pulse.detuning]) - {field.Uniform}

    if extra_sm:
        sm = extra_sm.pop()

        lattice_site_coefficients = GenerateLatticeSiteCoefficients(
            parallel_decoder=ahs_lattice_data.parallel_decoder
        ).emit(circuit)

        local_detuning = GeneratePiecewiseLinearChannel(
            sequence.rydberg, pulse.detuning, sm
        ).visit(circuit)

    return AHSComponents(
        lattice_data=ahs_lattice_data,
        global_detuning=global_detuning,
        global_amplitude=global_amplitude,
        global_phase=global_phase,
        local_detuning=local_detuning,
        lattice_site_coefficients=lattice_site_coefficients,
    )


def generate_quera_ir(
    ahs_components: AHSComponents, shots: int
) -> QuEraTaskSpecification:
    """7. generate quera ir

    This pass takes the AHS components and generates the QuEra IR.

    Args:
        ahs_components (AHSComponents): A collection of the AHS components
            generated for the given circuit.
        shots (int): Number of shots to run the circuit for.

    Returns:
        task_specification (QuEraTaskSpecification): QuEra IR for the given
            circuit.

    """
    import bloqade.submission.ir.task_specification as task_spec
    from bloqade.compiler.passes.hardware.units import (
        convert_time_units,
        convert_energy_units,
        convert_coordinate_units,
    )

    lattice = task_spec.Lattice(
        sites=list(
            map(
                convert_coordinate_units,
                ahs_components.lattice_data.sites,
            )
        ),
        filling=ahs_components.lattice_data.filling,
    )

    global_detuning = task_spec.GlobalField(
        times=list(map(convert_time_units, ahs_components.global_detuning.times)),
        values=list(map(convert_energy_units, ahs_components.global_detuning.values)),
    )

    local_detuning = None

    if ahs_components.lattice_site_coefficients is not None:
        local_detuning = task_spec.LocalField(
            times=list(map(convert_time_units, ahs_components.local_detuning.times)),
            values=list(
                map(convert_energy_units, ahs_components.local_detuning.values)
            ),
            lattice_site_coefficients=ahs_components.lattice_site_coefficients,
        )

    rabi_frequency_amplitude_field = task_spec.GlobalField(
        times=list(map(convert_time_units, ahs_components.global_amplitude.times)),
        values=list(map(convert_energy_units, ahs_components.global_amplitude.values)),
    )

    rabi_frequency_phase_field = task_spec.GlobalField(
        times=list(map(convert_time_units, ahs_components.global_phase.times)),
        values=ahs_components.global_phase.values,
    )

    detuning = task_spec.Detuning(
        global_=global_detuning,
        local=local_detuning,
    )

    rabi_frequency_amplitude = task_spec.RabiFrequencyAmplitude(
        global_=rabi_frequency_amplitude_field,
    )

    rabi_frequency_phase = task_spec.RabiFrequencyPhase(
        global_=rabi_frequency_phase_field,
    )

    rydberg = task_spec.RydbergHamiltonian(
        rabi_frequency_amplitude=rabi_frequency_amplitude,
        rabi_frequency_phase=rabi_frequency_phase,
        detuning=detuning,
    )

    effective_hamiltonian = task_spec.EffectiveHamiltonian(
        rydberg=rydberg,
    )

    return task_spec.QuEraTaskSpecification(
        nshots=shots,
        lattice=lattice,
        effective_hamiltonian=effective_hamiltonian,
    )


def generate_braket_ir(
    ahs_components: AHSComponents, shots: int
) -> BraketTaskSpecification:
    """7. generate braket ir

    This pass takes the AHS components and generates the Braket IR.

    Args:
        ahs_components (AHSComponents): A collection of the AHS components
            generated for the given circuit.
        shots (int): Number of shots to run the circuit for.

    Returns:
        task_specification (BraketTaskSpecification): Braket IR for the given
            circuit.

    """
    import braket.ir.ahs as ahs
    from bloqade.compiler.passes.hardware.units import (
        convert_time_units,
        convert_energy_units,
        convert_coordinate_units,
    )

    ahs_register = ahs.AtomArrangement(
        sites=list(map(convert_coordinate_units, ahs_components.lattice_data.sites)),
        filling=ahs_components.lattice_data.filling,
    )

    global_detuning_time_series = ahs.TimeSeries(
        times=list(map(convert_time_units, ahs_components.global_detuning.times)),
        values=list(map(convert_energy_units, ahs_components.global_detuning.values)),
    )

    local_detuning_time_series = None
    if ahs_components.lattice_site_coefficients is not None:
        local_detuning_time_series = ahs.TimeSeries(
            times=list(map(convert_time_units, ahs_components.local_detuning.times)),
            values=list(
                map(convert_energy_units, ahs_components.local_detuning.values)
            ),
        )

    amplitude_time_series = ahs.TimeSeries(
        times=list(map(convert_time_units, ahs_components.global_amplitude.times)),
        values=list(map(convert_energy_units, ahs_components.global_amplitude.values)),
    )

    phase_time_series = ahs.TimeSeries(
        times=list(map(convert_time_units, ahs_components.global_phase.times)),
        values=ahs_components.global_phase.values,
    )

    detuning = ahs.PhysicalField(
        time_series=global_detuning_time_series,
        pattern="uniform",
    )

    amplitude = ahs.PhysicalField(
        time_series=amplitude_time_series,
        pattern="uniform",
    )

    phase = ahs.PhysicalField(
        time_series=phase_time_series,
        pattern="uniform",
    )

    local_detuning = None
    if ahs_components.lattice_site_coefficients is not None:
        local_detuning = ahs.PhysicalField(
            time_series=local_detuning_time_series,
            pattern=ahs_components.lattice_site_coefficients,
        )

    driving_field = ahs.DrivingField(
        detuning=detuning,
        amplitude=amplitude,
        phase=phase,
    )

    shiftingFields = []
    if ahs_components.lattice_site_coefficients is not None:
        shiftingFields = [ahs.ShiftingField(magnitude=local_detuning)]

    program = ahs.Program(
        setup=ahs.Setup(ahs_register=ahs_register),
        hamiltonian=ahs.Hamiltonian(
            drivingFields=[driving_field],
            shiftingFields=shiftingFields,
        ),
    )

    return BraketTaskSpecification(nshots=shots, program=program)



from bloqade import start
import numpy as np
from bloqade.emulate.ir.space import Space
from bloqade.emulate.ir.emulator import Register
from bloqade.emulate.ir.atom_type import TwoLevelAtom, ThreeLevelAtom
from bloqade.emulate.ir.state_vector import _expt_two_body_op, _expt_one_body_op
from scipy.sparse import csc_array
from decimal import Decimal
from itertools import product
from functools import reduce
import pytest


def callback_single_atom(register, metadata, ham):
    density_op = np.array([[0.0, 0.0], [0, 1]])
    rabi_op = np.array([[0.0, -1.0j], [+1.0j, 0.0]])

    density = register.local_trace(density_op, 0)
    rabi_expt = register.local_trace(rabi_op, 0)

    exact_density = np.vdot(register.data, density_op.dot(register.data))
    exact_rabi_expt = np.vdot(register.data, rabi_op.dot(register.data))

    avg, std = ham.average_and_variance(register)

    np.testing.assert_almost_equal(exact_density, density)
    np.testing.assert_almost_equal(exact_rabi_expt, rabi_expt)


def test_expectation_value_single_atom():
    omega = 2 * np.pi
    (
        start.add_position((0, 0))
        .rydberg.rabi.amplitude.uniform.constant("omega", "run_time")
        .assign(omega=omega)
        .batch_assign(run_time=np.linspace(0, 2 * np.pi / omega, 11))
        .bloqade.python()
        .run_callback(callback=callback_single_atom)
    )


def callback_two_atom(register, *_):
    density_op = np.array([[0.0, 0.0], [0.0, 1.0]])
    rabi_op = np.array([[0.0, -1.0j], [0.0j, 0.0]])

    full_density_op_0 = np.kron(np.eye(2), density_op)
    full_density_op_1 = np.kron(density_op, np.eye(2))
    full_rabi_op_0 = np.kron(np.eye(2), rabi_op)
    full_rabi_op_1 = np.kron(rabi_op, np.eye(2))

    density_0 = register.local_trace(density_op, 0)
    density_1 = register.local_trace(density_op, 1)
    rabi_expt_0 = register.local_trace(rabi_op, 0)
    rabi_expt_1 = register.local_trace(rabi_op, 1)

    exact_density_0 = np.vdot(register.data, full_density_op_0.dot(register.data))
    exact_density_1 = np.vdot(register.data, full_density_op_1.dot(register.data))
    exact_rabi_expt_0 = np.vdot(register.data, full_rabi_op_0.dot(register.data))
    exact_rabi_expt_1 = np.vdot(register.data, full_rabi_op_1.dot(register.data))

    np.testing.assert_almost_equal(exact_density_0, density_0)
    np.testing.assert_almost_equal(exact_density_1, density_1)

    np.testing.assert_almost_equal(exact_rabi_expt_0, rabi_expt_0)
    np.testing.assert_almost_equal(exact_rabi_expt_1, rabi_expt_1)


def test_expection_value_two_atom():
    omega = 2 * np.pi
    (
        start.add_position((0, 0))
        .add_position((0, 6.1))
        .rydberg.rabi.amplitude.uniform.constant("omega", "run_time")
        .assign(omega=omega)
        .batch_assign(run_time=np.linspace(0, 2 * np.pi / omega, 11))
        .bloqade.python()
        .run_callback(callback=callback_two_atom)
    )


def callback_two_body(register, *_):
    plus_op = np.array([[0.0, 1.0], [0.0, 0.0]])
    minus_op = np.array([[0.0, 0.0], [1.0, 0.0]])

    two_body_operator = np.kron(minus_op, plus_op)
    # reverse order
    full_body_operator = reduce(np.kron, [plus_op, np.eye(2), minus_op])

    corr = register.local_trace(two_body_operator, (0, 2))

    expected_corr = np.vdot(register.data, full_body_operator.dot(register.data))

    np.testing.assert_almost_equal(corr, expected_corr)


def test_expectation_value_two_body():
    omega = 2 * np.pi
    (
        start.add_position((0, 0))
        .add_position((0, 6.1))
        .add_position((6.1, 6.1))
        .rydberg.rabi.amplitude.uniform.constant("omega", "run_time")
        .assign(omega=omega)
        .batch_assign(run_time=np.linspace(0, 2 * np.pi / omega, 11))
        .bloqade.python()
        .run_callback(callback=callback_two_body)
    )


def project_to_subspace(operator, configurations):
    from scipy.sparse import csr_matrix

    proj_shape = (configurations.size, operator.shape[0])
    data = np.ones_like(configurations)
    rows = np.arange(configurations.size)
    cols = configurations
    proj = csr_matrix((data, (rows, cols)), shape=proj_shape)

    if operator.ndim == 1:
        return proj @ operator
    else:
        return proj @ operator @ proj.T


@pytest.mark.parametrize(
    "atom_type, i", product([TwoLevelAtom, ThreeLevelAtom], [0, 1, 2])
)
def test_internals_single_body(atom_type, i):
    sites = [(0.0, 0.0), (0.0, 1.0), (0.0, 5.0)]

    space = Space.create(Register(atom_type, sites, Decimal("1.0")))

    op = np.random.normal(size=(atom_type.n_level, atom_type.n_level))
    ops_dict = {}
    ops_dict[space.n_atoms - i - 1] = op

    Id = np.eye(atom_type.n_level)
    op_full = reduce(np.kron, (ops_dict.get(site, Id) for site in range(space.n_atoms)))

    density_op_space = project_to_subspace(op_full, space.configurations)

    psi = np.random.normal(0, 1, size=space.size)
    psi /= np.linalg.norm(psi)

    # _expt_one_body_op()
    result = _expt_one_body_op.py_func(
        space.configurations, space.atom_type.n_level, psi, i, op
    )

    exact_result = np.vdot(psi, density_op_space.dot(psi))
    np.testing.assert_almost_equal(exact_result, result)


param_iter = (
    (atom_type, i, j)
    for atom_type, i, j in product([TwoLevelAtom, ThreeLevelAtom], range(4), range(4))
    if i != j
)


@pytest.mark.parametrize("atom_type, i, j", param_iter)
def test_internals_two_body(atom_type, i, j):
    sites = [(0.0, 0.0), (0.0, 1.0), (0.0, 5.0), (4.0, 0.0)]

    space = Space.create(Register(atom_type, sites, Decimal("1.0")))

    ops_dict = {}

    op_i = np.random.normal(size=(atom_type.n_level, atom_type.n_level))
    op_j = np.random.normal(size=(atom_type.n_level, atom_type.n_level))
    Id = np.eye(atom_type.n_level)

    ops_dict[space.n_atoms - i - 1] = op_i
    ops_dict[space.n_atoms - j - 1] = op_j

    sp_op = csc_array(np.kron(op_i, op_j))

    density_corr_full = reduce(
        np.kron, (ops_dict.get(site, Id) for site in range(space.n_atoms))
    )

    density_op_space = project_to_subspace(density_corr_full, space.configurations)

    psi = np.random.normal(0, 1, size=space.size)
    psi /= np.linalg.norm(psi)

    result = _expt_two_body_op.py_func(
        space.configurations,
        space.atom_type.n_level,
        psi,
        (i, j),
        sp_op.data,
        sp_op.indices,
        sp_op.indptr,
    )

    exact_result = np.vdot(psi, density_op_space.dot(psi))
    np.testing.assert_almost_equal(exact_result, result)



from .location import AtomArrangement, ParallelRegister, LocationInfo, ListOfLocations
from .bravais import (
    BoundedBravais,
    Chain,
    Square,
    Rectangular,
    Honeycomb,
    Triangular,
    Lieb,
    Kagome,
)


start = ListOfLocations()
"""
A Program starting point, alias of empty
[`ListOfLocations`][bloqade.ir.location.list.ListOfLocations].

- Next possible steps to build your program are:
- Specify which level coupling to address with:
    - `start.rydberg`: for [`Rydberg`][bloqade.builder.coupling.Rydberg]
        Level coupling
    - `start.hyperfine`: for [`Hyperfine`][bloqade.builder.coupling.Hyperfine]
        Level coupling
    - LOCKOUT: You cannot add atoms to your geometry after specifying level coupling.
- continue/start building your geometry with:
    - `start.add_position()`: to add atom(s) to current register. It will accept:
        - A single coordinate, represented as a tuple (e.g. `(5,6)`) with a value that
          can either be:
            - integers: `(5,6)`
            - floats: `(5.1, 2.5)`
            - strings (for later variable assignment): `("x", "y")`
            - [`Scalar`][bloqade.ir.scalar.Scalar] objects: `(2*cast("x"), 5+cast("y"))`
        - A list of coordinates, represented as a list of types mentioned previously.
        - A numpy array with shape (n, 2) where n is the total number of atoms
"""


__all__ = [
    "start",
    "AtomArrangement",
    "Chain",
    "Square",
    "Rectangular",
    "Honeycomb",
    "Triangular",
    "Lieb",
    "Kagome",
    "BoundedBravais",
    "ListOfLocations",
    "ParallelRegister",
    "LocationInfo",
]



from bloqade.compiler.codegen.hardware.lattice import AHSLatticeData
from bloqade.compiler.codegen.hardware.piecewise_linear import PiecewiseLinear
from bloqade.compiler.codegen.hardware.piecewise_constant import PiecewiseConstant
from beartype.typing import Optional, List
from pydantic.dataclasses import dataclass
from decimal import Decimal


@dataclass
class AHSComponents:
    lattice_data: AHSLatticeData
    global_detuning: PiecewiseLinear
    global_amplitude: PiecewiseLinear
    global_phase: PiecewiseConstant
    local_detuning: Optional[PiecewiseLinear]
    lattice_site_coefficients: Optional[List[Decimal]]



import inspect
from bloqade.compiler.analysis.python.waveform import WaveformScan
from bloqade.compiler.rewrite.python.waveform import NormalizeWaveformPython
from bloqade.compiler.codegen.python.waveform import CodegenPythonWaveform
from bloqade.compiler.rewrite.common.assign_variables import AssignBloqadeIR
from bloqade.factory import piecewise_linear, piecewise_constant
import bloqade.ir.control.waveform as wf
from unittest.mock import patch
import numpy as np
import numba
from bloqade import start


def f(t):
    return np.sin(2 * np.pi * t)


def test_waveform_scan():
    wf1 = wf.Poly([1, 2, 3], 4.0)
    wf2 = wf.Linear(0, 1, 4.0)
    wf3 = wf.Constant(1.0, 4.0)
    wf4 = wf.PythonFn.create(f, 4.0)

    for a_wf in [wf1, wf2, wf3, wf4]:
        scan = WaveformScan().scan(a_wf)

        assert scan.bindings == {a_wf: "__bloqade_var0"}
        assert scan.imports == {}

    wf5 = wf1 + wf2

    scan = WaveformScan().scan(wf5)
    assert scan.bindings == {
        wf5: "__bloqade_var0",
        wf1: "__bloqade_var1",
        wf2: "__bloqade_var2",
    }

    wf6 = 3.0 * wf1

    scan = WaveformScan().scan(wf6)
    assert scan.bindings == {
        wf6: "__bloqade_var0",
        wf1: "__bloqade_var1",
    }

    wf7 = -wf1

    scan = WaveformScan().scan(wf7)
    assert scan.bindings == {
        wf7: "__bloqade_var0",
        wf1: "__bloqade_var1",
    }

    wf8 = wf1[0:1]
    scan = WaveformScan().scan(wf8)
    assert scan.bindings == {
        wf8: "__bloqade_var0",
        wf1: "__bloqade_var1",
    }

    wf9 = wf1.append(wf2)
    scan = WaveformScan().scan(wf9)
    assert scan.bindings == {
        wf9: "__bloqade_var0",
        wf1: "__bloqade_var1",
        wf2: "__bloqade_var2",
    }

    wf10 = wf1 + wf2
    wf11 = wf10 + wf3
    scan = WaveformScan().scan(wf11)
    assert scan.bindings == {
        wf11: "__bloqade_var0",
        wf10: "__bloqade_var1",
        wf1: "__bloqade_var2",
        wf2: "__bloqade_var3",
        wf3: "__bloqade_var4",
    }


def test_waveform_normalize():
    wf1 = wf.Poly([1, 2, 3], 4.0)
    wf2 = wf.Linear(0, 1, 4.0)
    wf3 = wf.Constant(1.0, 4.0)
    wf4 = wf.PythonFn.create(f, 4.0)

    for a_wf in [
        wf1,
        wf2,
        wf3,
        wf4,
        wf1 + wf2,
        2.0 * wf1,
        -wf1,
        wf1[0:1],
        wf1.append(wf2),
    ]:
        assert NormalizeWaveformPython().visit(a_wf) == a_wf

    wf5 = wf4.sample(0.1, wf.Interpolation.Linear)

    times, values = wf5.samples()

    durations = np.diff(times)

    wf6 = piecewise_linear(durations.tolist(), values)

    assert NormalizeWaveformPython().visit(wf5) == wf6

    wf7 = wf4.sample(0.1, wf.Interpolation.Constant)

    times, values = wf5.samples()

    durations = np.diff(times)

    wf8 = piecewise_constant(durations.tolist(), values[:-1])

    assert NormalizeWaveformPython().visit(wf7) == wf8


@patch("bloqade.compiler.codegen.python.waveform.randint")
def test_python_codegen(randint):
    def func_def(wf):
        scan = WaveformScan().scan(wf)
        return CodegenPythonWaveform(scan).emit_func(wf, "test_func")[-1]

    def g(t, a, b):
        return a * np.sin(2 * np.pi * t) + b

    wf1 = wf.Poly([1, 2, 3], 3.0)
    wf2 = wf.Linear(0, 1, 4.0)
    wf3 = wf.Constant(1.0, 4.0)
    wf4 = wf.PythonFn.create(f, 4.0)
    wf5 = wf.PythonFn.create(g, 4.0)

    wf6 = AssignBloqadeIR(dict(a=1, b=2)).emit(wf5)

    randint.side_effect = [0, 1, 2, 3, 4]

    assert (func_def(wf1)) == (
        "def test_func(time):\n"
        "    if time > 3.0:\n"
        "        return 0\n"
        "    __bloqade_var0 = 1 + 2 * (time) ** 1 + 3 * (time) ** 2\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf2)) == (
        "def test_func(time):\n"
        "    if time > 4.0:\n"
        "        return 0\n"
        "    __bloqade_var0 = 0.25 * (time) + 0\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf3)) == (
        "def test_func(time):\n"
        "    if time > 4.0:\n"
        "        return 0\n"
        "    __bloqade_var0 = 1.0\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf4)) == (
        "def test_func(time):\n"
        "    if time > 4.0:\n"
        "        return 0\n"
        "    __bloqade_var0 = __bloqade_waveform_0(time)\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf6)) == (
        "def test_func(time):\n"
        "    if time > 4.0:\n"
        "        return 0\n"
        "    __bloqade_var0 = __bloqade_waveform_1(time, a = 1, b = 2)\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf1 + wf2)) == (
        "def test_func(time):\n"
        "    if time > 4.0:\n"
        "        return 0\n"
        "    __bloqade_var1 = 1 + 2 * (time) ** 1 + 3 * (time) ** 2\n"
        "    __bloqade_var2 = 0.25 * (time) + 0\n"
        "    __bloqade_var0 = __bloqade_var1 + __bloqade_var2 if time "
        "< 3.0 else __bloqade_var2\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf2 + wf1)) == (
        "def test_func(time):\n"
        "    if time > 4.0:\n"
        "        return 0\n"
        "    __bloqade_var1 = 0.25 * (time) + 0\n"
        "    __bloqade_var2 = 1 + 2 * (time) ** 1 + 3 * (time) ** 2\n"
        "    __bloqade_var0 = __bloqade_var1 + __bloqade_var2 if time "
        "< 3.0 else __bloqade_var1\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf2 + wf3)) == (
        "def test_func(time):\n"
        "    if time > 4.0:\n"
        "        return 0\n"
        "    __bloqade_var1 = 0.25 * (time) + 0\n"
        "    __bloqade_var2 = 1.0\n"
        "    __bloqade_var0 = __bloqade_var1 + __bloqade_var2\n"
        "    return __bloqade_var0"
    )

    assert (func_def(2.0 * wf1)) == (
        "def test_func(time):\n"
        "    if time > 3.0:\n"
        "        return 0\n"
        "    __bloqade_var1 = 1 + 2 * (time) ** 1 + 3 * (time) ** 2\n"
        "    __bloqade_var0 = 2.0 * __bloqade_var1\n"
        "    return __bloqade_var0"
    )

    assert (func_def(-wf1)) == (
        "def test_func(time):\n"
        "    if time > 3.0:\n"
        "        return 0\n"
        "    __bloqade_var1 = 1 + 2 * (time) ** 1 + 3 * (time) ** 2\n"
        "    __bloqade_var0 = -__bloqade_var1\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf1[1:2])) == (
        "def test_func(time):\n"
        "    if time > 1:\n"
        "        return 0\n"
        "    __bloqade_var1 = 1 + 2 * (time + 1) ** 1 + 3 * (time + 1) ** 2\n"
        "    __bloqade_var0 = __bloqade_var1\n"
        "    return __bloqade_var0"
    )

    assert (func_def(wf1.append(wf2))) == (
        "def test_func(time):\n"
        "    if time > 7.0:\n"
        "        return 0\n"
        "    if time < 3.0:\n"
        "        __bloqade_var1 = 1 + 2 * (time - 0) ** 1 + 3 * (time - 0) ** 2\n"
        "        __bloqade_var0 = __bloqade_var1\n"
        "    elif time <= 7.0:\n"
        "        __bloqade_var2 = 0.25 * (time - 3.0) + 0\n"
        "        __bloqade_var0 = __bloqade_var2\n"
        "    else:\n"
        "        __bloqade_var0 = 0\n"
        "    return __bloqade_var0"
    )

    scan = WaveformScan().scan(wf1)
    func = CodegenPythonWaveform(scan).compile(wf1)

    assert isinstance(func, numba.core.registry.CPUDispatcher)
    assert func.__name__ == "__bloqade_waveform_2"

    scan = WaveformScan().scan(wf1)
    func = CodegenPythonWaveform(scan, jit_compiled=False).compile(wf1)

    assert inspect.isfunction(func)
    assert func.__name__ == "__bloqade_waveform_3"


def test_interpret_vs_python_vs_numba():
    def phase_function(t):
        return np.sin(t)

    def collect_callback(state, metadata, hamiltonian):
        return state.data.copy()

    program = (
        start.add_position([(0, 0), (0, 6)])
        .rydberg.detuning.uniform.piecewise_linear([0.1, 0.8, 0.1], [-10, -10, 20, 20])
        .amplitude.uniform.piecewise_linear([0.1, 0.8, 0.1], [0, 15.7, 15.7, 0])
        .phase.location(0, 1.0)
        .fn(phase_function, 1.0)
        .detuning.scale([0.1, 1.0])
        .piecewise_constant([0.1, 0.3, 0.1], [0, 1, 2])
        .poly([2, -1, 3, -4], 0.5)
    )

    interp_results = program.bloqade.python().run_callback(
        collect_callback, waveform_runtime="interpret"
    )

    python_results = program.bloqade.python().run_callback(
        collect_callback, waveform_runtime="python"
    )

    numba_results = program.bloqade.python().run_callback(
        collect_callback, waveform_runtime="numba"
    )

    for interp_result, python_result, numba_result in zip(
        interp_results, python_results, numba_results
    ):
        assert np.allclose(interp_result, python_result)
        assert np.allclose(interp_result, numba_result)


if __name__ == "__main__":
    test_python_codegen()



from functools import cached_property
from pydantic.dataclasses import dataclass
from typing import Dict, List, Tuple, Union
from decimal import Decimal


ParamType = Union[Decimal, List[Decimal]]


@dataclass(frozen=True)
class ScalarArg:
    name: str


@dataclass(frozen=True)
class VectorArg:
    name: str


@dataclass(frozen=True)
class Params:
    n_sites: int
    static_params: Dict[str, ParamType]
    batch_params: List[Dict[str, ParamType]]
    args_list: Tuple[Union[ScalarArg, VectorArg], ...]

    @cached_property
    def num_args(self) -> int:
        num_args = 0
        for arg in self.args_list:
            if isinstance(arg, VectorArg):
                # expect n_sites args for vector arguments
                num_args += self.n_sites
            else:
                num_args += 1

        return num_args

    @cached_property
    def arg_names(self) -> Tuple[str]:
        return tuple([arg.name for arg in self.args_list])

    def parse_args(self, *flattened_args) -> Dict[str, Decimal]:
        if len(flattened_args) != self.num_args:
            raise ValueError(
                f"Expected {self.num_args} arguments, got {len(flattened_args)}."
            )

        args = []
        i = 0
        for arg in self.args_list:
            # if arg is a vector, then we need to unpack the next n_sites args
            if isinstance(arg, VectorArg):
                vec = list(map(Decimal, map(str, flattened_args[i : i + self.n_sites])))
                args.append(vec)
                i += self.n_sites
            else:
                args.append(Decimal(str(flattened_args[i])))
                i += 1

        return dict(zip(self.arg_names, args))

    def batch_assignments(self, *args) -> List[Dict[str, ParamType]]:
        flattened_args = self.parse_args(*args)
        if len(self.batch_params) == 0:
            # handle case where there are no batch params
            return [flattened_args]
        else:
            return [{**flattened_args, **batch} for batch in self.batch_params]

    def __str__(self):
        out = ""
        out += "> Static params:\n"
        for var, litrl in self.static_params.items():
            out += f" :: {var} \n      => {litrl}\n"

        out += "\n> Batch params:\n"
        for lid, pair in enumerate(self.batch_params):
            out += f"- batch {lid}:\n"
            for var, litrl in pair.items():
                out += f"   :: {var}\n       => {litrl}\n"

        out += "\n> Arguments:\n"
        out += "  " + repr(self.args_list)

        return out



from bloqade import cast
from bloqade.ir import scalar
from bloqade.compiler.rewrite.common.assign_to_literal import AssignToLiteral
from bloqade.compiler.rewrite.common.assign_variables import AssignBloqadeIR
from decimal import Decimal


def test_assign_to_literal():
    a = cast("a")
    b = cast("b")
    c = cast("c")

    expr = (a - b) * c / 2.0

    a = Decimal("1.0")
    b = Decimal("2.0")
    c = Decimal("3.0")

    assigned_expr = AssignBloqadeIR(dict(a=a, b=b, c=c)).visit(expr)

    assert AssignToLiteral().visit(assigned_expr) == scalar.Div(
        scalar.Mul(
            scalar.Add(scalar.Literal(a), scalar.Negative(scalar.Literal(b))),
            scalar.Literal(c),
        ),
        scalar.Literal(2.0),
    )



import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
import bloqade.ir.control.waveform as waveform
import bloqade.ir.scalar as scalar
from bloqade.ir.visitor import BloqadeIRTransformer


class AddPadding(BloqadeIRTransformer):
    def __init__(
        self,
        level_couplings=None,
        field_names=None,
        spatial_modulations=None,
    ):
        self.level_couplings = level_couplings
        self.field_names = field_names
        self.spatial_modulations = spatial_modulations
        self.duration = None

    def add_waveform_padding(self, wf, duration):
        if wf is None:
            return waveform.Constant(0, duration)

        diff_duration = duration - wf.duration
        if diff_duration == scalar.Literal(0):
            return wf

        if isinstance(wf, waveform.AlignedWaveform):
            raise NotImplementedError("AlignedWaveform not implemented yet")
            # if isinstance(wf.value, waveform.Side):
            #     value = scalar.var(f"__record_value_{hash(wf) ^ hash(wf.value)}__")
            #     wf = wf.record(value, wf.value)
            #     padding = waveform.Constant(value, diff_duration)
            # else:
            #     value = wf.value

            # padding = waveform.Constant(wf.value, diff_duration)

            # if wf.alignment is waveform.Alignment.Left:
            #     return wf.append(padding)
            # else:
            #     return padding.append(wf)
        else:
            padding = waveform.Constant(0, diff_duration)
            return wf.append(padding)

    def get_empty_field(self, duration):
        empty_wf = waveform.Constant(0, duration)
        empty_f = field.Field({sm: empty_wf for sm in self.spatial_modulations})
        return empty_f

    def get_empty_pulse(self, duration):
        fields = {}
        for fn, spatial_modulations in self.field_names.items():
            self.spatial_modulations = spatial_modulations
            fields[fn] = self.get_empty_field(duration)

        return pulse.Pulse(fields)

    #######################
    # Visitor definitions #
    #######################

    def visit_sequence_Sequence(self, node: sequence.Sequence) -> sequence.Sequence:
        pulses = {}
        for lc, field_names in self.level_couplings.items():
            self.field_names = field_names

            if lc in node.pulses:
                p = node.pulses[lc]
                diff_duration = node.duration - p.duration

                p = self.visit(p)
                if diff_duration != scalar.Literal(0):
                    p = p.append(self.get_empty_pulse(diff_duration))

            else:
                p = self.get_empty_pulse(node.duration)

            pulses[lc] = p

        return sequence.Sequence(pulses)

    def visit_pulse_Pulse(self, node: pulse.Pulse) -> pulse.Pulse:
        self.pulse_duration = node.duration

        fields = {}
        for fn, spatial_modulations in self.field_names.items():
            self.spatial_modulations = spatial_modulations
            if fn in node.fields:
                fields[fn] = self.visit(node.fields[fn])
            else:  # no need to visit empty field
                fields[fn] = self.get_empty_field(self.pulse_duration)

        return pulse.Pulse(fields)

    def visit_field_Field(self, node: field.Field) -> field.Field:
        duration = (
            self.pulse_duration if self.pulse_duration is not None else node.duration
        )
        drives = {}

        for sm in self.spatial_modulations:
            drives[sm] = self.visit(
                self.add_waveform_padding(node.drives.get(sm), duration)
            )

        return field.Field(drives)

    def visit_waveform_Add(self, node: waveform.Add) -> waveform.Add:
        left = self.visit(node.left)
        right = self.visit(node.right)

        if left.duration == right.duration:
            return left + right
        else:
            duration = left.duration.max(right.duration)
            return self.add_waveform_padding(
                left, duration
            ) + self.add_waveform_padding(right, duration)



from bloqade.builder.typing import LiteralType
from bloqade.ir.visitor import BloqadeIRVisitor
import bloqade.ir.control.waveform as waveform
from typing import Dict
from decimal import Decimal


class AssignmentScan(BloqadeIRVisitor):
    def __init__(self, assignments: Dict[str, LiteralType] = {}):
        self.assignments = dict(assignments)

    def visit_waveform_Record(self, node: waveform.Record):
        self.visit(node.waveform)
        duration = node.waveform.duration(**self.assignments)
        var = node.var

        if node.side is waveform.Side.Right:
            value = node.waveform.eval_decimal(duration, **self.assignments)
        else:
            value = node.waveform.eval_decimal(Decimal(0), **self.assignments)

        self.assignments[var.name] = value

    def scan(self, node) -> Dict[str, LiteralType]:
        self.visit(node)
        return self.assignments



from bloqade.ir.visitor import BloqadeIRTransformer
import bloqade.ir.scalar as scalar
from bloqade.ir.control import waveform


class AssignToLiteral(BloqadeIRTransformer):
    """Transform all assigned variables to literals."""

    def visit_scalar_AssignedVariable(self, node: scalar.AssignedVariable):
        return scalar.Literal(node.value)

    def visit_waveform_PythonFn(self, node: waveform.PythonFn):
        return node  # skip these nodes



from decimal import Decimal
from beartype.typing import Tuple


def convert_time_units(time: Decimal) -> Decimal:
    return time * Decimal("1e-6")


def convert_energy_units(energy: Decimal) -> Decimal:
    return energy * Decimal("1e6")


def convert_coordinate_units(
    length: Tuple[Decimal, Decimal]
) -> Tuple[Decimal, Decimal]:
    return (length[0] * Decimal("1e-6"), length[1] * Decimal("1e-6"))



import bloqade.ir.control.waveform as waveform
import bloqade.ir.control.field as field
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.sequence as sequence
from bloqade import var

# from bloqade.ir.scalar import var

from bloqade.compiler.analysis.common.scan_channels import ScanChannels


def generate_waveforms():
    wf1 = waveform.Constant("v1", "t1")
    wf2 = waveform.Constant("v2", "t2")
    wf3 = waveform.Constant("v3", "t3")
    wf4 = waveform.Constant(4.0, "t4")

    return wf1, wf2, wf3, wf4


def generate_spatiamodulations():
    sm1 = field.UniformModulation()
    sm2 = field.RunTimeVector("mask1")
    sm3 = field.AssignedRunTimeVector("mask2", list(range(4)))
    sm4 = field.ScaledLocations.create(
        {field.Location(0): "m0", field.Location(2): "m2"}
    )

    return sm1, sm2, sm3, sm4


def generate_fields():
    wf1, wf2, wf3, wf4 = generate_waveforms()
    sm1, sm2, sm3, sm4 = generate_spatiamodulations()

    f1 = field.Field({sm1: wf1})
    f2 = field.Field({sm3: wf3})
    f3 = field.Field({sm2: wf2, sm3: wf3})
    f4 = field.Field({sm2: wf2, sm4: wf4, sm1: wf1})

    return f1, f2, f3, f4


def test_scan_channels_fields():
    sm1, sm2, sm3, sm4 = generate_spatiamodulations()
    f1, f2, f3, f4 = generate_fields()

    assert ScanChannels().scan(f1) == {sm1}
    assert ScanChannels().scan(f2) == {sm3}
    assert ScanChannels().scan(f3) == {sm2, sm3}
    assert ScanChannels().scan(f4) == {sm1, sm2, sm4}


def generate_pulses():
    f1, f2, f3, f4 = generate_fields()

    p1 = pulse.Pulse({pulse.rabi.amplitude: f1})
    p2 = pulse.Pulse({pulse.rabi.phase: f2})
    p3 = pulse.Pulse({pulse.rabi.amplitude: f4, pulse.detuning: f3})

    p4 = p1.append(p2)
    p5 = p3[: var("t")]
    p6 = p4.append(p3)

    return p1, p2, p3, p4, p5, p6


def test_scan_channels_pulses():
    sm1, sm2, sm3, sm4 = generate_spatiamodulations()
    p1, p2, p3, p4, p5, p6 = generate_pulses()

    assert ScanChannels().scan(p1) == {pulse.rabi.amplitude: {sm1}}
    assert ScanChannels().scan(p3) == {
        pulse.rabi.amplitude: {sm1, sm2, sm4},
        pulse.detuning: {sm2, sm3},
    }
    assert ScanChannels().scan(p4) == {
        pulse.rabi.amplitude: {sm1},
        pulse.rabi.phase: {sm3},
    }
    assert ScanChannels().scan(p5) == {
        pulse.rabi.amplitude: {sm1, sm2, sm4},
        pulse.detuning: {sm2, sm3},
    }
    assert ScanChannels().scan(p6) == {
        pulse.rabi.amplitude: {sm1, sm2, sm4},
        pulse.rabi.phase: {sm3},
        pulse.detuning: {sm2, sm3},
    }


def generate_sequences():
    p1, p2, p3, p4, p5, p6 = generate_pulses()

    s1 = sequence.Sequence({sequence.hyperfine: p1})
    s2 = sequence.Sequence({sequence.rydberg: p2})
    s3 = s1.append(s2)
    s4 = s3[: var("t")]

    return s1, s2, s3, s4


def test_scan_channel_sequence():
    sm1, sm2, sm3, sm4 = generate_spatiamodulations()
    s1, s2, s3, s4 = generate_sequences()

    assert ScanChannels().scan(s1) == {
        sequence.hyperfine: {
            pulse.rabi.amplitude: {sm1},
        }
    }
    assert ScanChannels().scan(s2) == {
        sequence.rydberg: {
            pulse.rabi.phase: {sm3},
        }
    }
    assert ScanChannels().scan(s3) == {
        sequence.hyperfine: {
            pulse.rabi.amplitude: {sm1},
        },
        sequence.rydberg: {
            pulse.rabi.phase: {sm3},
        },
    }
    assert ScanChannels().scan(s4) == {
        sequence.hyperfine: {
            pulse.rabi.amplitude: {sm1},
        },
        sequence.rydberg: {
            pulse.rabi.phase: {sm3},
        },
    }



from functools import reduce
from bloqade.ir.control import waveform, pulse, sequence, field
from bloqade.ir import analog_circuit

from bloqade.ir.visitor import BloqadeIRVisitor

from beartype.typing import List
from beartype import beartype
from pydantic.dataclasses import dataclass
from bisect import bisect_right, bisect_left
from decimal import Decimal


@dataclass(frozen=True)
class PiecewiseConstant:
    times: List[Decimal]
    values: List[Decimal]

    def eval(self, time):
        if time < 0 or time > self.times[-1]:
            return Decimal("0")

        i = bisect_right(self.times[1:], time)

        return self.values[i]

    def slice(self, start_time, stop_time) -> "PiecewiseConstant":
        start_time = Decimal(str(start_time))
        stop_time = Decimal(str(stop_time))

        if start_time == stop_time:
            return PiecewiseConstant(
                [Decimal(0.0), Decimal(0.0)], [Decimal(0.0), Decimal(0.0)]
            )

        start_index = bisect_left(self.times, start_time)
        stop_index = bisect_left(self.times, stop_time)

        if start_time == self.times[start_index]:
            if stop_time == self.times[stop_index]:
                absolute_times = list(self.times[start_index : stop_index + 1])
                values = list(self.values[start_index : stop_index + 1])
            else:
                absolute_times = self.times[start_index:stop_index] + [stop_time]
                values = self.values[start_index:stop_index] + [self.values[stop_index]]
        else:
            if stop_time == self.times[stop_index]:
                absolute_times = [start_time] + self.times[start_index : stop_index + 1]
                values = [self.values[start_index - 1]] + self.values[
                    start_index : stop_index + 1
                ]
            else:
                absolute_times = (
                    [start_time] + self.times[start_index:stop_index] + [stop_time]
                )
                values = (
                    [self.values[start_index - 1]]
                    + self.values[start_index:stop_index]
                    + [self.values[stop_index]]
                )

        values[-1] = values[-2]

        return PiecewiseConstant([time - start_time for time in absolute_times], values)

    @staticmethod
    def append(
        left: "PiecewiseConstant", right: "PiecewiseConstant"
    ) -> "PiecewiseConstant":
        new_times = left.times + [time + left.times[-1] for time in right.times[1:]]
        new_values = left.values[:-1] + right.values
        return PiecewiseConstant(
            times=new_times,
            values=new_values,
        )


class GeneratePiecewiseConstantChannel(BloqadeIRVisitor):
    valid_nodes = {
        waveform.Constant,
        waveform.Linear,
        waveform.Poly,
        waveform.Sample,
        waveform.Add,
        waveform.Append,
        waveform.Slice,
        waveform.Negative,
        waveform.Scale,
        field.Field,
        pulse.Pulse,
        pulse.NamedPulse,
        pulse.Slice,
        pulse.Append,
        sequence.Sequence,
        sequence.NamedSequence,
        sequence.Slice,
        sequence.Append,
        analog_circuit.AnalogCircuit,
    }

    @beartype
    def __init__(
        self,
        level_coupling: sequence.LevelCoupling,
        field_name: pulse.FieldName,
        spatial_modulations: field.SpatialModulation,
    ):
        self.level_coupling = level_coupling
        self.field_name = field_name
        self.spatial_modulations = spatial_modulations

    def visit_waveform_Constant(self, node: waveform.Constant) -> PiecewiseConstant:
        return PiecewiseConstant(
            [Decimal(0), node.duration()], [node.value(), node.value()]
        )

    def visit_waveform_Linear(self, node: waveform.Linear) -> PiecewiseConstant:
        return PiecewiseConstant(
            [Decimal(0), node.duration()], [node.start(), node.start()]
        )

    def visit_waveform_Poly(self, node: waveform.Poly) -> PiecewiseConstant:
        duration = node.duration()
        start = node.eval_decimal(0)

        return PiecewiseConstant(
            [Decimal(0), duration],
            [start, start],
        )

    def visit_waveform_Sample(self, node: waveform.Sample) -> PiecewiseConstant:
        times, values = node.samples()
        values[-1] = values[-2]

        return PiecewiseConstant(times, values)

    def visit_waveform_Add(self, node: waveform.Add) -> PiecewiseConstant:
        left = self.visit(node.left)
        right = self.visit(node.right)

        times = sorted(list(set(left.times + right.times)))
        values = [left.eval(t) + right.eval(t) for t in times]

        return PiecewiseConstant(times, values)

    def visit_waveform_Append(self, node: waveform.Append) -> PiecewiseConstant:
        return reduce(PiecewiseConstant.append, map(self.visit, node.waveforms))

    def visit_waveform_Slice(self, node: waveform.Slice) -> PiecewiseConstant:
        pwl = self.visit(node.waveform)
        return pwl.slice(node.start(), node.stop())

    def visit_waveform_Negative(self, node: waveform.Negative) -> PiecewiseConstant:
        pwl = self.visit(node.waveform)
        return PiecewiseConstant(pwl.times, [-v for v in pwl.values])

    def visit_waveform_Scale(self, node: waveform.Scale) -> PiecewiseConstant:
        pwl = self.visit(node.waveform)
        return PiecewiseConstant(pwl.times, [node.scalar() * v for v in pwl.values])

    def visit_field_Field(self, node: field.Field) -> PiecewiseConstant:
        return self.visit(node.drives[self.spatial_modulations])

    def visit_pulse_Pulse(self, node: pulse.Pulse) -> PiecewiseConstant:
        return self.visit(node.fields[self.field_name])

    def visit_pulse_NamedPulse(self, node: pulse.NamedPulse) -> PiecewiseConstant:
        return self.visit(node.pulse)

    def visit_pulse_Slice(self, node: pulse.Slice) -> PiecewiseConstant:
        return self.visit(node.pulse).slice(node.start(), node.stop())

    def visit_pulse_Append(self, node: pulse.Append) -> PiecewiseConstant:
        return reduce(PiecewiseConstant.append, map(self.visit, node.pulses))

    def visit_sequence_Sequence(self, node: sequence.Sequence) -> PiecewiseConstant:
        return self.visit(node.pulses[self.level_coupling])

    def visit_sequence_NamedSequence(
        self, node: sequence.NamedSequence
    ) -> PiecewiseConstant:
        return self.visit(node.sequence)

    def visit_sequence_Slice(self, node: sequence.Slice) -> PiecewiseConstant:
        return self.visit(node.sequence).slice(node.start(), node.stop())

    def visit_sequence_Append(self, node: sequence.Append) -> PiecewiseConstant:
        return reduce(PiecewiseConstant.append, map(self.visit, node.sequences))

    def visit_analog_circuit_AnalogCircuit(
        self, node: analog_circuit.AnalogCircuit
    ) -> PiecewiseConstant:
        return self.visit(node.sequence)

    # add another visit method for the new type
    # every visitor method should return a PiecewiseConstant object
    # Then we can use the PiecewiseConstant object to combine the
    # resulting waveforms into a single waveform using the AST node
    # to determine which transformation to apply, e.g. add, scale, etc.
    def visit(self, node) -> PiecewiseConstant:
        if type(node) not in self.valid_nodes:
            raise TypeError(
                f"Expected one of {self.valid_nodes}, got {type(node)} instead."
            )

        return super().visit(node)



from bloqade.ir.visitor import BloqadeIRVisitor
from bloqade.ir.control import waveform, pulse, sequence
from beartype.typing import Union


class CheckSlices(BloqadeIRVisitor):
    def check_slice(self, node: Union[waveform.Slice, pulse.Slice, sequence.Slice]):
        start_time = node.start()
        stop_time = node.stop()

        duration = node.waveform.duration()

        if start_time < 0:
            raise ValueError(f"Start time {start_time} is negative")

        if stop_time < 0:
            raise ValueError(f"Stop time {stop_time} is negative")

        if start_time > stop_time:
            raise ValueError(
                f"Start time {start_time} is greater than stop time {stop_time}"
            )

        if stop_time > duration:
            raise ValueError(
                f"Stop time {stop_time} is greater than waveform duration {duration}"
            )

    def visit_waveform_Slice(self, node: waveform.Slice):
        self.visit(node.waveform)
        self.check_slice(node)

    def visit_pulse_Slice(self, node: pulse.Slice):
        self.visit(node.pulse)
        self.check_slice(node)

    def visit_sequence_Slice(self, node: sequence.Slice):
        self.visit(node.sequence)
        self.check_slice(node)



from .channels import ValidateChannels
from .piecewise_linear import ValidatePiecewiseLinearChannel
from .piecewise_constant import ValidatePiecewiseConstantChannel


__all__ = [
    "ValidateChannels",
    "ValidatePiecewiseLinearChannel",
    "ValidatePiecewiseConstantChannel",
]



from beartype.typing import Any
import bloqade.ir.control.sequence as sequence
from bloqade.ir.visitor import BloqadeIRVisitor


class IsHyperfineSequence(BloqadeIRVisitor):
    def __init__(self):
        self.is_hyperfine = False

    def generic_visit(self, node: Any) -> Any:
        # skip visiting children if we already know there are hyperfine pulses
        if self.is_hyperfine:
            return

        super().generic_visit(node)

    def visit_sequence_Sequence(self, node: sequence.Sequence) -> Any:
        self.is_hyperfine = self.is_hyperfine or sequence.hyperfine in node.pulses

    def emit(self, node) -> bool:
        self.visit(node)
        return self.is_hyperfine
